# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("dhcpcsvc.dll")
prototypes = \
    {
        #
        'DhcpCApiInitialize': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Version"]),
        #
        'DhcpCApiCleanup': SimTypeFunction([], SimTypeBottom(label="Void")),
        #
        'DhcpRequestParams': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"Flags": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "nBytesData": SimTypeInt(signed=False, label="UInt32")}, name="DHCPCAPI_CLASSID", pack=False, align=None), offset=0), SimStruct({"nParams": SimTypeInt(signed=False, label="UInt32"), "Params": SimTypePointer(SimStruct({"Flags": SimTypeInt(signed=False, label="UInt32"), "OptionId": SimTypeInt(signed=False, label="UInt32"), "IsVendor": SimTypeInt(signed=True, label="Int32"), "Data": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "nBytesData": SimTypeInt(signed=False, label="UInt32")}, name="DHCPAPI_PARAMS", pack=False, align=None), offset=0)}, name="DHCPCAPI_PARAMS_ARRAY", pack=False, align=None), SimStruct({"nParams": SimTypeInt(signed=False, label="UInt32"), "Params": SimTypePointer(SimStruct({"Flags": SimTypeInt(signed=False, label="UInt32"), "OptionId": SimTypeInt(signed=False, label="UInt32"), "IsVendor": SimTypeInt(signed=True, label="Int32"), "Data": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "nBytesData": SimTypeInt(signed=False, label="UInt32")}, name="DHCPAPI_PARAMS", pack=False, align=None), offset=0)}, name="DHCPCAPI_PARAMS_ARRAY", pack=False, align=None), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Flags", "Reserved", "AdapterName", "ClassId", "SendParams", "RecdParams", "Buffer", "pSize", "RequestIdStr"]),
        #
        'DhcpUndoRequestParams': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Flags", "Reserved", "AdapterName", "RequestIdStr"]),
        #
        'DhcpRegisterParamChange': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"Flags": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "nBytesData": SimTypeInt(signed=False, label="UInt32")}, name="DHCPCAPI_CLASSID", pack=False, align=None), offset=0), SimStruct({"nParams": SimTypeInt(signed=False, label="UInt32"), "Params": SimTypePointer(SimStruct({"Flags": SimTypeInt(signed=False, label="UInt32"), "OptionId": SimTypeInt(signed=False, label="UInt32"), "IsVendor": SimTypeInt(signed=True, label="Int32"), "Data": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "nBytesData": SimTypeInt(signed=False, label="UInt32")}, name="DHCPAPI_PARAMS", pack=False, align=None), offset=0)}, name="DHCPCAPI_PARAMS_ARRAY", pack=False, align=None), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Flags", "Reserved", "AdapterName", "ClassId", "Params", "Handle"]),
        #
        'DhcpDeRegisterParamChange': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Flags", "Reserved", "Event"]),
        #
        'DhcpRemoveDNSRegistrations': SimTypeFunction([], SimTypeInt(signed=False, label="UInt32")),
        #
        'DhcpGetOriginalSubnetMask': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["sAdapterName", "dwSubnetMask"]),
        #
        'McastApiStartup': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Version"]),
        #
        'McastApiCleanup': SimTypeFunction([], SimTypeBottom(label="Void")),
        #
        'McastGenUID': SimTypeFunction([SimTypePointer(SimStruct({"ClientUID": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ClientUIDLength": SimTypeInt(signed=False, label="UInt32")}, name="MCAST_CLIENT_UID", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pRequestID"]),
        #
        'McastEnumerateScopes': SimTypeFunction([SimTypeShort(signed=False, label="UInt16"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"ScopeCtx": SimStruct({"ScopeID": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None"), "Interface": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None"), "ServerID": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None")}, name="MCAST_SCOPE_CTX", pack=False, align=None), "LastAddr": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None"), "TTL": SimTypeInt(signed=False, label="UInt32"), "ScopeDesc": SimTypeBottom(label="UNICODE_STRING")}, name="MCAST_SCOPE_ENTRY", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["AddrFamily", "ReQuery", "pScopeList", "pScopeLen", "pScopeCount"]),
        #
        'McastRequestAddress': SimTypeFunction([SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimStruct({"ClientUID": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ClientUIDLength": SimTypeInt(signed=False, label="UInt32")}, name="MCAST_CLIENT_UID", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"ScopeID": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None"), "Interface": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None"), "ServerID": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None")}, name="MCAST_SCOPE_CTX", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"LeaseStartTime": SimTypeInt(signed=True, label="Int32"), "MaxLeaseStartTime": SimTypeInt(signed=True, label="Int32"), "LeaseDuration": SimTypeInt(signed=False, label="UInt32"), "MinLeaseDuration": SimTypeInt(signed=False, label="UInt32"), "ServerAddress": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None"), "MinAddrCount": SimTypeShort(signed=False, label="UInt16"), "AddrCount": SimTypeShort(signed=False, label="UInt16"), "pAddrBuf": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="MCAST_LEASE_REQUEST", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"LeaseStartTime": SimTypeInt(signed=True, label="Int32"), "LeaseEndTime": SimTypeInt(signed=True, label="Int32"), "ServerAddress": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None"), "AddrCount": SimTypeShort(signed=False, label="UInt16"), "pAddrBuf": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="MCAST_LEASE_RESPONSE", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["AddrFamily", "pRequestID", "pScopeCtx", "pAddrRequest", "pAddrResponse"]),
        #
        'McastRenewAddress': SimTypeFunction([SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimStruct({"ClientUID": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ClientUIDLength": SimTypeInt(signed=False, label="UInt32")}, name="MCAST_CLIENT_UID", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"LeaseStartTime": SimTypeInt(signed=True, label="Int32"), "MaxLeaseStartTime": SimTypeInt(signed=True, label="Int32"), "LeaseDuration": SimTypeInt(signed=False, label="UInt32"), "MinLeaseDuration": SimTypeInt(signed=False, label="UInt32"), "ServerAddress": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None"), "MinAddrCount": SimTypeShort(signed=False, label="UInt16"), "AddrCount": SimTypeShort(signed=False, label="UInt16"), "pAddrBuf": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="MCAST_LEASE_REQUEST", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"LeaseStartTime": SimTypeInt(signed=True, label="Int32"), "LeaseEndTime": SimTypeInt(signed=True, label="Int32"), "ServerAddress": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None"), "AddrCount": SimTypeShort(signed=False, label="UInt16"), "pAddrBuf": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="MCAST_LEASE_RESPONSE", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["AddrFamily", "pRequestID", "pRenewRequest", "pRenewResponse"]),
        #
        'McastReleaseAddress': SimTypeFunction([SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimStruct({"ClientUID": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ClientUIDLength": SimTypeInt(signed=False, label="UInt32")}, name="MCAST_CLIENT_UID", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"LeaseStartTime": SimTypeInt(signed=True, label="Int32"), "MaxLeaseStartTime": SimTypeInt(signed=True, label="Int32"), "LeaseDuration": SimTypeInt(signed=False, label="UInt32"), "MinLeaseDuration": SimTypeInt(signed=False, label="UInt32"), "ServerAddress": SimUnion({"IpAddrV4": SimTypeInt(signed=False, label="UInt32"), "IpAddrV6": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="<anon>", label="None"), "MinAddrCount": SimTypeShort(signed=False, label="UInt16"), "AddrCount": SimTypeShort(signed=False, label="UInt16"), "pAddrBuf": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="MCAST_LEASE_REQUEST", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["AddrFamily", "pRequestID", "pReleaseRequest"]),
    }

lib.set_prototypes(prototypes)



import angr

import logging

l = logging.getLogger(name=__name__)


class strtok_r(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, str_ptr, delim_ptr, save_ptr, str_strlen=None, delim_strlen=None):
        if self.state.libc.simple_strtok:
            malloc = angr.SIM_PROCEDURES["libc"]["malloc"]
            token_ptr = self.inline_call(malloc, self.state.libc.strtok_token_size).ret_expr
            r = self.state.solver.If(
                self.state.solver.Unconstrained("strtok_case", self.state.arch.bits) == 0,
                token_ptr,
                self.state.solver.BVV(0, self.state.arch.bits),
            )
            self.state.libc.strtok_heap.append(token_ptr)
            return r
        else:
            strstr = angr.SIM_PROCEDURES["libc"]["strstr"]
            strlen = angr.SIM_PROCEDURES["libc"]["strlen"]

            l.debug("Doin' a strtok_r!")
            l.debug("... geting the saved state")

            saved_str_ptr = self.state.memory.load(
                save_ptr, self.state.arch.bytes, endness=self.state.arch.memory_endness
            )
            start_ptr = self.state.solver.If(str_ptr == 0, saved_str_ptr, str_ptr)

            l.debug("... getting the lengths")
            str_strlen = self.inline_call(strlen, start_ptr) if str_strlen is None else str_strlen
            delim_strlen = self.inline_call(strlen, delim_ptr) if delim_strlen is None else delim_strlen

            l.debug("... STRTOK SLOW PATH (symbolic-length delimiteter and/or string)")
            l.debug("... calling strstr")
            where = self.inline_call(
                strstr, start_ptr, delim_ptr, haystack_strlen=str_strlen, needle_strlen=delim_strlen
            )
            write_length = self.state.solver.If(where.ret_expr != 0, delim_strlen.ret_expr, 0)
            write_content = self.state.solver.BVV(0, delim_strlen.max_null_index * 8)

            # do a symbolic write (we increment the limit because of the possibility that the write target is 0,
            # in which case the length will be 0, anyways)
            l.debug("... doing the symbolic write")
            self.state.memory.store(
                where.ret_expr,
                write_content,
                size=write_length,
                strategy=["symbolic_nonzero", "any"],
                limit=str_strlen.max_null_index + 1,
            )

            l.debug("... creating the return address")
            new_start = write_length + where.ret_expr
            new_state = self.state.solver.If(new_start != 0, new_start, start_ptr)

            l.debug("... saving the state")
            self.state.memory.store(save_ptr, new_state, endness=self.state.arch.memory_endness)

            l.debug("... done")
            return new_start



#!/usr/bin/env python3
__package__ = __package__ or "tests.state_plugins.inspect"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ...common import bin_location, broken


test_location = os.path.join(bin_location, "tests")

target_addrs = {"i386": [0x080485C9], "x86_64": [0x4006ED], "ppc": [0x1000060C], "armel": [0x85F0], "mips": [0x4009FC]}

avoid_addrs = {
    "i386": [0x080485DD, 0x08048564],
    "x86_64": [0x4006AA, 0x4006FD],
    "ppc": [0x10000644, 0x1000059C],
    "armel": [0x86F8, 0x857C],
    "mips": [0x400A10, 0x400774],
}

corrupt_addrs = {
    "i386": [0x80486B6, b"bO\xcc", lambda s: s.memory.store(s.regs.esp, s.regs.eax)],
    "x86_64": [0x400742, b"\xd4&\xb0[\x41", lambda s: s.registers.store("rdx", 8)],
    "ppc": [0x100006B8, b"\x05\xad\xc2\xea", lambda s: s.registers.store("r5", 8)],
    "armel": [0x8678, b"\xbdM\xec3", lambda s: s.registers.store("r2", 8)],
    "mips": [0x400918, b"[\xf8\x96@"[::-1], lambda s: s.registers.store("a2", 8)],
}


class TestSyscallOverride(unittest.TestCase):
    def _run_fauxware_override(self, arch):
        # auto_load_libs can't be disabled as the test fails
        p = angr.Project(os.path.join(test_location, arch, "fauxware"), use_sim_procedures=False)
        s = p.factory.full_init_state()

        def overwrite_str(state):
            state.posix.get_fd(1).write_data(b"HAHA\0")

        queued_syscall_returns = []
        queued_syscall_returns.append(None)  # let the mmap run
        queued_syscall_returns.append(overwrite_str)  # prompt for username
        queued_syscall_returns.append(0)  # username read
        queued_syscall_returns.append(0)  # newline read
        # queued_syscall_returns.append(0) # prompt for password -- why isn't this called?
        queued_syscall_returns.append(None)  # password input
        queued_syscall_returns.append(0)  # password \n input

        def syscall_hook(state):
            if not state.inspect.simprocedure.is_syscall:
                return
            try:
                f = queued_syscall_returns.pop(0)
                if f is None:
                    return
                state.inspect.simprocedure_result = f(state) if callable(f) else f
            except IndexError:
                return

        s.inspect.make_breakpoint("simprocedure", s.inspect.BP_BEFORE, action=syscall_hook)

        results = p.factory.simulation_manager(thing=s).explore(find=target_addrs[arch], avoid=avoid_addrs[arch])
        stdin = results.found[0].posix.dumps(0)
        assert b"SOSNEAKY" == stdin
        stdout = results.found[0].posix.dumps(1)
        assert b"HAHA\0" == stdout

    def test_fauxware_override_x86_64(self):
        self._run_fauxware_override("x86_64")

    def test_fauxware_override_i386(self):
        self._run_fauxware_override("i386")

    @broken
    def test_fauxware_override_ppc(self):
        self._run_fauxware_override("ppc")

    @broken
    def test_fauxware_override_armel(self):
        self._run_fauxware_override("armel")

    @broken
    def test_fauxware_override_mips(self):
        self._run_fauxware_override("mips")


if __name__ == "__main__":
    unittest.main()
